from datetime import datetime
from typing import Any, Dict, Optional, Union
from uuid import UUID

from odp.util.cheapdantic import BaseModel


class FileMetadataDto(BaseModel):
    """File Metadata Model."""

    name: str
    mime_type: Optional[str] = None
    dataset: Optional[UUID] = None
    metadata: Dict[str, Union[bool, int, str]] = {}
    geo_location: Optional[Any] = None
    size_bytes: Optional[int] = None
    checksum: Optional[str] = None
    created_time: Optional[datetime] = None
    modified_time: Optional[datetime] = None
    deleted_time: Optional[datetime] = None

    def __init__(self, name: str, **kwargs):
        if name.startswith("/"):
            raise ValueError("name cannot start with '/'. Absolute paths are not allowed.")
        super().__init__(name=name, **kwargs)



from typing import Dict, Optional
from uuid import UUID

from odp.dto.validators import validate_resource_name
from odp.util.cheapdantic import BaseModel, Field


class MetadataDto(BaseModel):
    """Resource manifest metadata"""

    name: str
    """Resource name. Must consist of alphanumeric characters, dashes or underscores and must start
    with an alphanumeric character"""

    display_name: Optional[str] = None
    """Human-friendly name"""

    description: Optional[str] = None
    """Resource description"""

    uuid: Optional[UUID] = None
    """System-assigned unique identifier"""

    labels: Dict = Field(default_factory=dict)
    """Resource labels"""

    owner: Optional[UUID] = None

    def __init__(self, **kwargs):
        kwargs["name"] = validate_resource_name(kwargs.get("name"))
        super().__init__(**kwargs)



from typing import Optional

from odp.util.cheapdantic import BaseModel


class License(BaseModel):
    """Data license information"""

    name: str
    """License name. Can be set to `Proprietary` for proprietary licenses"""

    href: Optional[str] = None
    """HREF to license text"""

    full_text: Optional[str] = None
    """Full license text"""



from typing import Any, Dict, List, Optional, Set

from ..common.contact_info import ContactInfo
from ..resource import ResourceDto, ResourceSpec
from ..resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from ..validators import validate_at_least_one_not_null, validate_doi
from .distribution import Distribution
from .observable import ObservableDto
from odp.util.cheapdantic import BaseModel, Field


class Citation(BaseModel):
    """Citation information"""

    cite_as: Optional[str] = None
    """Directions on how to cite the dataset"""

    doi: Optional[str] = None

    def __init__(self, **kwargs):
        kwargs = validate_at_least_one_not_null(kwargs)
        kwargs["doi"] = validate_doi(kwargs.get("doi"))

        super().__init__(**kwargs)


class Attribute(BaseModel):
    """Dataset attribute"""

    name: str
    """Attribute name. This can be a column name in a table, a dimension in an array, etc."""

    description: Optional[str] = None
    """Attribute description"""

    traits: list[str]
    """List of traits. Traits are used to describe the attribute in more detail.

    Traits are based on Microsoft Common Data Model (CDM) traits. See the [CDM documentation]
    (https://learn.microsoft.com/en-us/common-data-model/sdk/trait-concepts-and-use-cases#what-are-traits)
    for more information.
    """


class DatasetSpec(ResourceSpec):
    distribution: Optional[Distribution] = None
    """Information on how the dataset was distributed"""

    storage_class: str
    """Storage class qualified name"""

    storage_controller: Optional[str] = None
    """Storage controller qualified name"""

    data_collection: Optional[str] = None
    """Data collection qualified name"""

    maintainer: ContactInfo
    """Active maintainer information"""

    citation: Optional[Citation] = None
    """Citation information"""

    documentation: List[str] = Field(default_factory=list)
    """Links to any relevant documentation"""

    attributes: List[Attribute] = Field(default_factory=list)
    """Dataset attributes"""

    facets: Optional[Dict[str, Any]] = None
    """Facets for the dataset"""

    observables: List[ObservableDto] = Field(default_factory=list)
    """Observables for the dataset"""

    tags: Set[str] = Field(default_factory=set)


class DatasetDto(ResourceDto):
    _kind: str = "catalog.hubocean.io/dataset"
    _version: str = "v1alpha3"

    spec: DatasetSpec


DEFAULT_RESOURCE_TYPE_REGISTRY.add(DatasetDto)



import io
import json
import logging
import typing
from functools import partial
from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Union

import pandas as pd
import pyarrow as pa
import requests

import odp.tabular_v2.client as old
from odp.tabular_v2.client.table_raw import Raw
from odp.tabular_v2.util import vars_to_json
from odp.util.cheapdantic import BaseModel

if TYPE_CHECKING:
    from odp.tabular_v2.client import Cursor


class TableStats(BaseModel):
    num_rows: int
    """number of rows in the table"""

    size: int
    """size of the table in bytes, including metadata and schema"""


# temporary: migrate old API to the new client
def _request_to_new_client(
    cli,  # new client instance
    path: str,
    data: Union[Dict, bytes, Iterator[bytes], io.IOBase, None] = None,
    params: Optional[Dict] = None,
    headers: Optional[Dict] = None,
) -> old.Client.Response:
    req = requests.Request(
        method="POST",
        url=cli.base_url + path,
        params=params,
        headers=headers,
        data=data,
    )
    res = cli._request(req)
    return old.Client.Response(res)


class Table:
    def __init__(self, impl, table_id: str):
        self._id = table_id

        # unify new and old client _request() method
        import odp.new_client as new

        if isinstance(impl, new.Client):
            self._impl = partial(_request_to_new_client, impl)
        elif isinstance(impl, old.Client):
            self._impl = impl._request
        else:
            raise ValueError(f"Unexpected implementation type: {type(impl)}")

        self._tx = None

    @property
    def raw(self) -> "Raw":
        return Raw(self)

    def stats(self) -> TableStats:
        """
        Get the table statistics, such as the number of rows, size, etc.
        @return: a TableStats object containing the statistics, including:
                 - num_rows: the number of rows in the table
                 - size: the size of the table in bytes, including metadata and schema
        """
        res = self._impl(
            path="/api/table/v2/stats",
            params={
                "table_id": self._id,
            },
        )
        return TableStats(**res.json())

    def schema(self) -> typing.Optional[pa.Schema]:
        try:
            empty = list(self._select(inner_query='"fetch" == "schema"'))
        except FileNotFoundError:
            return None
        assert len(empty) == 1
        assert empty[0].num_rows == 0
        return empty[0].schema

    def alter(self, schema: pa.Schema, from_names: dict = {}):
        """
        perform a schema change, re-ingesting all the data in the table with the new schema
        from_names is a dictionary mapping new names to old names, used to rename fields or duplicate them
        """
        j = json.dumps(from_names).encode("utf-8")
        # create a stream with only the schema
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, schema)
        w.write_batch(
            pa.RecordBatch.from_pylist([], schema),
            {"rename": j},
        )
        w.close()

        res = self._impl(
            path="/api/table/v2/sdk/alter",
            params={
                "table_id": self._id,
            },
            data=buf.getvalue(),  # send the schema using pa.ipc
        )
        return res.json()

    def drop(self):
        """
        drop the table data and schema
        this operation is irreversible
        @return:
        """
        try:
            res = self._impl(
                path="/api/table/v2/drop",
                params={
                    "table_id": self._id,
                },
            ).json()
            logging.info("dropped %s: %s", self._id, res)
        except FileNotFoundError:
            logging.info("table %s does not exist", self._id)

    def _validate_parquet_schema(self, schema: pa.Schema):
        compatible_types = {
            pa.types.is_integer,
            pa.types.is_floating,
            pa.types.is_boolean,
            pa.types.is_string,
            pa.types.is_binary,
            pa.types.is_date,
            pa.types.is_timestamp,
            pa.types.is_decimal,
            pa.types.is_time,
        }
        for field in schema:
            if not any(check_function(field.type) for check_function in compatible_types):
                raise ValueError(f"Incompatible type for parquet detected: {field.name} ({field.type})")

    def create(self, schema: pa.Schema):
        """
        set the table schema using the given pyarrow schema
        fields might contains metadata which will be used internally:
        * index: the field should be used to partition the data
        * isGeometry: the field is a geometry (wkt for string, wkb for binary)
        @param schema: pyarrow.Schema
        @raise FileExistsError if the schema is already set
        @return:
        """
        self._validate_parquet_schema(schema)
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, schema)
        w.write_batch(pa.RecordBatch.from_pylist([], schema=schema))
        w.close()

        self._impl(
            path="/api/table/v2/sdk/create",
            params={
                "table_id": self._id,
            },
            data=buf.getvalue(),
        ).json()

    def aggregate(
        self,
        by: str = '"TOTAL"',
        query: str = "",
        aggr: Union[dict, None] = None,
        timeout: float = 30.0,
        vars: Union[dict, list, None] = None,
    ) -> pd.DataFrame:
        """
        aggregate the data after the optional `query` filter
        the paramater `by` is used to determine the key for the aggregation, and can be an expression.
        the optional `aggr` specify which fields need to be aggregated, and how
        If not specified, the fields with metadata "aggr" will be used
        a single DataFrame will be returned, with the index set to the key used for aggregation
        """
        schema = self.schema()
        if schema is None:
            raise FileNotFoundError(f"Table {self._id} does not exist")

        if aggr is None:
            aggr = {}
            for field in schema:
                if field.metadata and b"aggr" in field.metadata:
                    aggr[field.name] = field.metadata[b"aggr"].decode()

        tot_func = {
            "*": "sum",
        }
        for field, a_type in aggr.items():
            if a_type == "mean" or a_type == "avg":
                tot_func[field + "_sum"] = "sum"
                tot_func[field + "_count"] = "sum"
            elif a_type == "sum":
                tot_func[field + "_sum"] = "sum"
            elif a_type == "min":
                tot_func[field + "_min"] = "min"
            elif a_type == "max":
                tot_func[field + "_max"] = "max"
            elif a_type == "count":
                tot_func[field + "_count"] = "sum"
            else:
                raise ValueError(f"unknown aggregation type: {a_type}")

        total: Union[pd.DataFrame, None] = None
        for b in self._select(type="aggregate", by=by, inner_query=query, timeout=timeout, aggr=aggr, vars=vars):
            df: pd.DataFrame = b.to_pandas()
            # logging.warning("PARTIAL:\n%s", df)
            if total is None:
                total = df
            else:
                total = pd.concat([total, df], ignore_index=True)
                total = total.groupby("").agg(tot_func).reset_index()
        if total is None:
            return pd.DataFrame()

        for field, a_type in aggr.items():
            logging.info("field: %s, type: %s", field, a_type)
            if a_type == "mean" or a_type == "avg":
                total[field] = total[field + "_sum"] / total[field + "_count"]
                total.drop(columns=[field + "_sum", field + "_count"], inplace=True)
            elif a_type in "sum":
                total[field] = total[field + "_sum"]
                total.drop(columns=[field + "_sum"], inplace=True)
            elif a_type == "min":
                total[field] = total[field + "_min"]
                total.drop(columns=[field + "_min"], inplace=True)
            elif a_type == "max":
                total[field] = total[field + "_max"]
                total.drop(columns=[field + "_max"], inplace=True)
            elif a_type == "count":
                total[field] = total[field + "_count"]
                total.drop(columns=[field + "_count"], inplace=True)
            else:
                raise ValueError(f"unknown aggregation type: {a_type}")

        total = total.set_index("")
        # logging.info("TOTAL:\n%s", total)
        return total

    def _query_cursor(
        self,
        query: str = "",
        cols: Optional[List[str]] = None,
        vars: Union[dict, list, None] = None,
        stream_ttl: float = 30.0,
        type: str = "select",
        tx_id: str = "",
    ) -> "Cursor":
        def scanner(scanner_cursor: str) -> Iterator[pa.RecordBatch]:
            logging.info("selecting cursor=%s, query=%s", scanner_cursor, query)
            for b in self._select(
                tx=tx_id,
                type=type,
                inner_query=query,
                cols=cols,
                vars=vars,
                cursor=scanner_cursor,
                timeout=stream_ttl,
            ):
                # logging.info("got %d rows, decoding...", b.num_rows)
                yield b

        from odp.tabular_v2.client import Cursor

        schema = self.schema()
        if cols:
            invalid_cols = [col for col in cols if col not in schema.names]
            if invalid_cols:
                raise ValueError(f"Invalid columns: {invalid_cols}. Available columns: {schema.names}")
            schema = pa.schema([schema.field(col) for col in cols])
        return Cursor(scanner=scanner, schema=schema)

    def select(
        self,
        query: str = "",
        cols: Optional[List[str]] = None,
        vars: Union[dict, list, None] = None,
        timeout: float = 30.0,
    ) -> "Cursor":
        """
        fetch data from the underling table

        for row in tab.select("age > 18").rows():
            print(row)

        you can use bind variables, especially if you need to use date/time objects:

        for row in tab.select("age > $age", vars={"age": 18}).rows():
            print(row)

        and limits which columns you want to retrieve:

        for row in tab.select("age > 18", cols=["name", "age"]).rows():
            print(row)

        The object returned is a cursor, which can be scanned by rows, batches, pages, pandas dataframes, etc.

        you can check the documentation of the Cursor for more information
        """
        return self._query_cursor(query, cols, vars, timeout, "select")

    def __enter__(self):
        if self._tx:
            raise ValueError("already in a transaction")

        res = self._impl(
            path="/api/table/v2/begin",
            params={
                "table_id": self._id,
            },
        ).json()
        from odp.tabular_v2.client.table_tx import Transaction

        self._tx = Transaction(self, res["tx_id"])
        return self._tx

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            logging.warning("aborting transaction %s", self._tx._id)
            # try:
            #    self._impl(
            #        path="/api/table/v2/rollback",
            #        params={
            #            "table_id": self._id,
            #            "tx_id": self._tx._id,
            #        },
            #    )
            # except Exception as e:
            #    logging.error("ignored: rollback failed: %s", e)
        else:
            self._tx.flush()
            self._impl(
                path="/api/table/v2/commit",
                params={
                    "table_id": self._id,
                    "tx_id": self._tx._id,
                },
            )
        self._tx = None

    # used as a filter in Cursor, encode in tx
    def _select(
        self,
        tx: str = "",
        type: str = "select",
        inner_query: str = "",
        aggr: Optional[dict] = None,
        cols: Optional[List[str]] = None,
        vars: Union[Dict, List, None] = None,
        by: Optional[str] = None,
        cursor: Union[str, None] = "",
        timeout: float = 30.0,
    ) -> Iterator[pa.RecordBatch]:
        # t0 = time.perf_counter()
        while cursor is not None:
            res = self._impl(
                path="/api/table/v2/sdk/" + type,
                params={
                    "table_id": self._id,
                    "tx_id": tx,
                },
                data={
                    "query": str(inner_query) if inner_query else None,
                    "cols": cols,
                    "cursor": cursor,
                    "aggr": aggr,
                    "by": by,
                    "vars": vars_to_json(vars),
                    "timeout": timeout,
                },
            )
            cursor = None
            reader = res.reader()
            r = pa.ipc.RecordBatchStreamReader(reader)
            for bm in r.iter_batches_with_custom_metadata():
                if bm.custom_metadata and b"error" in bm.custom_metadata:
                    raise ValueError("server error: %s" % bm.custom_metadata[b"error"].decode())

                if bm.custom_metadata:
                    if b"cursor" in bm.custom_metadata:
                        cursor = bm.custom_metadata[b"cursor"].decode()
                        logging.info("response is partially processed with cursor %s", cursor)
                    if b"stats" in bm.custom_metadata:
                        logging.info("stats: %s", bm.custom_metadata[b"stats"].decode())

                # logging.info("got batch with %d rows", bm.batch.num_rows)
                yield bm.batch

    def _insert_batch(
        self,
        data: pa.RecordBatch,
        tx: str = "",
    ):
        schema = self.schema()
        buf = io.BytesIO()
        w = pa.ipc.RecordBatchStreamWriter(buf, schema)
        w.write_batch(data)
        w.close()

        self._impl(
            path="/api/table/v2/sdk/insert",
            params={
                "table_id": self._id,
                "tx_id": tx,
            },
            data=buf.getvalue(),
        ).json()



import logging
import time
from typing import Callable, Iterator

import pandas as pd
import pyarrow as pa


class CursorException(Exception):
    """Raised when the client is required to connect again with the given cursor to fetch more data"""

    def __init__(self, cursor: str):
        self.cursor = cursor


class Cursor:
    """
    A Cursor, which reads data from a scanner and allow the user to iterate over in different flavors
    It internally uses a scanner to get the data, and perform multiple requests if a cursor is returned

    it's created by the Table.select() method
    """

    def __init__(
        self,
        scanner: Callable[[str], Iterator[pa.RecordBatch]],
        schema: pa.Schema,
    ):
        self.scanner = scanner
        self.schema = schema

    def dataframes(self) -> Iterator[pd.DataFrame]:
        """
        Allow the client to iterate over pyarrow Record Batches:

            for batch in tab.select().batches():
                print(batch.num_rows)
        """
        for b in self.batches():
            yield b.to_pandas()

    def all(self, max_rows: int = 1_000_000, max_time: float = 60.0):
        """
        attempt to fetch all the data from the server as a single table
        raises TimeoutError if it takes too long or too many rows
        """
        batches = []
        t0 = time.perf_counter()
        rows = 0
        for b in self.batches():
            rows += b.num_rows
            if rows > max_rows or time.perf_counter() - t0 > max_time:
                raise OverflowError("too many rows")
            if time.perf_counter() - t0 > max_time:
                raise TimeoutError("fetching all rows is taking too long")
            batches.append(b)

        class FlatList:
            def __init__(self, tab: pa.Table):
                self._tab = tab

            def table(self) -> pa.Table:
                return self._tab

            def dataframe(self) -> pd.DataFrame:
                return self._tab.to_pandas()

            def rows(self) -> list:
                return self._tab.to_pylist()

        if len(batches) == 0:
            return FlatList(pa.Table.from_pylist([], schema=self.schema))
        return FlatList(pa.Table.from_batches(batches))

    def batches(self) -> Iterator[pa.RecordBatch]:
        """
        Allow the client to iterate over pandas DataFrames:

            for batch in tab.select().dataframes():
                print(len(batch))
        """
        t0 = time.perf_counter()
        cursor = ""
        ct = 0
        rows = 0
        while True:
            try:
                for b in self.scanner(cursor):
                    ct += 1
                    rows += b.num_rows
                    if b.num_rows > 0:
                        yield b
            except CursorException as e:
                cursor = e.cursor
                continue
            break
        logging.info("got %d batches with %d rows in %.2fs", ct, rows, time.perf_counter() - t0)

    def rows(self) -> Iterator[dict]:
        """
        Allow the client to iterate over each rows as a python dictionary:

            for row in tab.select().rows():
                print(row)
        """
        for b in self.batches():
            for row in b.to_pylist():
                yield row

    def pages(self, size: int = 0) -> Iterator[list[dict]]:
        """
        Allow the client to iterate over pages of data, where each page is a list of python dictionaries.

            for page in tab.select().pages(1_000):
                print(len(page))

        If no size is specified, it will make a page for each chunk of data retrieved from the server
        """
        if size < 1:  # page based on what we get
            for b in self.batches():
                yield b.to_pylist()
            return

        # page based on page_size
        buf: list[dict] = []
        for b in self.batches():
            buf.extend(b.to_pylist())
            while len(buf) >= size:
                yield buf[:size]
                buf = buf[size:]
        if len(buf) > 0:
            yield buf



from abc import ABC
from datetime import datetime
from typing import Optional, Union
from uuid import UUID

from .metadata import MetadataDto
from .resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from .resource_status import ResourceStatus
from .validators import validate_resource_kind, validate_resource_version
from odp.util.cheapdantic import BaseModel


class ResourceSpec(BaseModel, ABC):
    class Config:
        extra = "allow"


class ResourceDto(BaseModel):
    """Resource Data Transmission Object (DTO) representing a resource manifest"""

    kind: str = None
    """kind is the kind of the resource."""

    version: str = None
    """version is the version of the resource."""

    metadata: MetadataDto
    """metadata is the metadata of the resource."""

    status: Optional[ResourceStatus] = None
    """status is the status of the resource."""

    spec: ResourceSpec
    """Resource spec"""

    def __init__(self, **kwargs):
        kwargs["kind"] = self._populate_kind(kwargs.get("kind"))
        kwargs["version"] = self._populate_version(kwargs.get("version"))
        super().__init__(**kwargs)

    @classmethod
    def create(
        cls,
        kind: str,
        version: str,
        metadata: MetadataDto,
        spec: ResourceSpec,
        status: Optional[ResourceStatus] = None,
    ) -> "ResourceDto":
        promoted_cls = DEFAULT_RESOURCE_TYPE_REGISTRY.get(kind, version, ResourceDto)
        return promoted_cls(kind=kind, version=version, metadata=metadata, spec=spec, status=status)

    def _populate_kind(cls, v) -> str:
        kind = cls.get_kind()
        if not kind:
            kind = v
        elif v and v != kind:
            raise ValueError(f"Invalid kind '{v}' for resource '{cls.__name__}' - expected '{kind}'")
        return validate_resource_kind(kind)

    def _populate_version(cls, v) -> str:
        version = cls.get_version()
        if not version:
            version = v
        elif v and v != version:
            raise ValueError(f"Invalid version '{v}' for resource '{cls.__name__}' - expected '{version}")
        return validate_resource_version(version)

    def populate_status(self, created_by: UUID):
        cls = DEFAULT_RESOURCE_TYPE_REGISTRY.get(self.kind, self.version, None)

        if cls and cls != type(self):
            cls.populate_status(self, created_by)
        else:
            if self.status is not None:
                raise ValueError("Resource status already set")

            self.status = ResourceStatus.create(created_by)

    def update_status(self, updated_by: UUID):
        cls = DEFAULT_RESOURCE_TYPE_REGISTRY.get(self.kind, self.version, None)

        if cls and cls != type(self):
            cls.update_status(self, updated_by)
        else:
            if self.status is None:
                raise ValueError("Resource status not set")

            self.status.num_updates += 1
            self.status.updated_time = datetime.now()
            self.status.updated_by = updated_by

    @property
    def qualified_name(self) -> str:
        """Return the qualified name of the resource

        The qualified name of resource is the concatenation of the resource kind and name, separated by a slash.
        """
        return f"{self.kind}/{self.metadata.name}"

    @property
    def uuid(self) -> Optional[UUID]:
        """Return the UUID of the resource"""
        return self.metadata.uuid

    @classmethod
    def get_kind(cls) -> str:
        """Return the kind of the resource"""
        return getattr(cls, "_kind", None)

    @classmethod
    def get_version(cls) -> str:
        """Return the version of the resource"""
        return getattr(cls, "_version", None)

    def get_ref(self) -> Union[UUID, str]:
        """Get a valid reference to the resource

        Returns:
            The resource UUID if it is set, the qualified name otherwise
        """
        return self.uuid or self.qualified_name



import re
from typing import Any, Dict, Mapping, Optional, Union

import jsonschema

RX_RESOURCE_NAME = re.compile(r"[a-zA-Z0-9][a-zA-Z0-9\-_.]*")
RX_RESOURCE_KIND = re.compile(r"^(?:[a-zA-Z0-9][a-zA-Z0-9\-_\.]*)\/(?:[a-zA-Z0-9][a-zA-Z0-9\-_\.]*)$")
RX_RESOURCE_VERSION = re.compile(r"^v[0-9]+(?:(?:alpha|beta)[0-9]+)?$")
RX_DOI = r"^(?:doi:\/\/)?(10\.[0-9]{4,9}(?:\.[0-9]+)*\/(?:(?![%\"#?\s])\S)+)$"


def validate_resource_version(val: str) -> str:
    if not RX_RESOURCE_VERSION.fullmatch(val):
        raise ValueError(f"Invalid resource version: {val}")
    return val


def validate_resource_kind(val: str) -> str:
    if not RX_RESOURCE_KIND.fullmatch(val):
        raise ValueError(f"Invalid resource kind: {val}")
    return val


def validate_resource_name(val: str) -> str:
    if not RX_RESOURCE_NAME.fullmatch(val):
        raise ValueError(f"Invalid resource component: {val}")
    return val


def validate_doi(val: Optional[str]) -> Optional[str]:
    if val is not None and not re.fullmatch(RX_DOI, val):
        raise ValueError(f"Invalid DOI: {val}")
    return val


def validate_json_schema(v: Union[bool, Mapping[str, Any]]) -> Union[bool, Mapping[str, Any]]:
    try:
        jsonschema.Draft4Validator.check_schema(v)
    except jsonschema.SchemaError as e:
        raise ValueError(f"Schema is invalid: {e.message}") from e
    return v


def validate_at_least_one_not_null(values: Dict[Any, Any]) -> Dict[Any, Any]:
    if not any(values.values()):
        raise ValueError(f"One of the fields must be set: {list(values.keys())}")

    return values
